#timecheck.tcl
#
# The program checks the consistency of the time signature
# read from usuls.dat, with the delta increments in the
# txt file.

#To run this program type:
#tclsh timecheck.tcl
#followed by <enter>

# textsrc indicates where to find the text files.
global textsrc
set textsrc txt/


proc loadusuls_v3 {} {
    global usulstype
    set inhandle [open usuls_v3.txt]
    gets $inhandle line
    gets $inhandle line
    while {![eof $inhandle]} {
        gets $inhandle line
        set linelist [split $line \t]
        set usulname [lindex $linelist 0]
        set usulstype($usulname) $line
    }
    close $inhandle
}

loadusuls_v3




proc loadusulinfo {} {
    global usulinfo
    puts "loadusulinfo"
    set inhandle [open edited_usuls.dat]
    #set inhandle [open usuls.dat]
    while {![eof $inhandle]} {
        gets $inhandle line
        set linelist [split $line \t]
        set filename [lindex $linelist 0]
        #  puts $usulname
        set usulinfo($filename) [lindex $linelist 1]
    }
    close $inhandle
}




global usulinfo
loadusulinfo





# Divides n1/d1 by n2/d2 and eliminates
# common factors using Euclid's algorithm.
proc fracdivision {n1 d1 n2 d2} {
    global triplet
    set n [expr $n1 * $d2]
    set d [expr $d1 * $n2]
    set m [euclids_algorithm $n $d]
    set n [lindex $m 0]
    set d [lindex $m 1]
    if {[expr $d % 3] == 0} {
        incr triplet
        set d [expr $d/3]
        if {$n > 1} {set n [expr $n/2]
                   } else {set d [expr $d*2]}
        if {$triplet >= 3} {incr triplet -3}
        }
    if {$n == 1 && $d == 1} {return ""}
    if {$d == 1} {return $n}
    if {$d == 2 && $n == 1} {return /}
    return $n/$d
}

proc euclids_algorithm {a b} {
    #eliminate common factor in fraction a/b
    if {$a > $b} {
        set n $a; set m $b} else {
        set n $b; set m $a}
    while {$m != 0} {
        set t [expr $n % $m]
        set n $m
        set m $t
    }
    set a [expr $a/$n]
    set b [expr $b/$n]
    return "$a $b"
}



# Splits the file name into various parts that
# are separated by a double dash --.

proc separate_by_double_dash {name} {
set namelist {}
set i 0
while {[set j [string first -- $name $i]] > -1} {
  set j0 [expr $j - 1]
  set s [string range $name $i $j0]
  lappend namelist $s
  set i [expr $j + 2]
  }
  lappend namelist [string range $name $i end]
  return $namelist
}





# mcsplit --
#
#   Splits a string based using another string
#
# Arguments:
#   str       string to split into pieces
#   splitStr  substring
#   mc        magic character that must not exist in the orignal string.
#             Defaults to the NULL character.  Must be a single character.
# Results:
#   Returns a list of strings
#
proc mcsplit "str splitStr {mc {\x00}}" {
    return [split [string map [list $splitStr $mc] $str] $mc]
    }




#global filedatainfo


set xrefnum 1

proc analyze_abc_file {fullfilename} {
    global textsrc
    global filedata
    global usulinfo
    global triplet
    global xrefnum
    global usulstype
 
    set ldenom 16
    set triplet 0
    set l [string length $fullfilename]
    incr l -5
    set filename [string range $fullfilename 0 $l]

    set filenamelist [separate_by_double_dash $filename]
    set filedata [mcsplit $filename "--"]
    set makamdata [lindex $filedata 0]
    set formdata [lindex $filedata 1]
    set usulname [lindex $filedata 2]
    set rhythminfo $usulinfo($filename)

    if {[info exist usulstype($usulname)]} {
          set usultxt $usulstype($usulname)
          #set usullist [split $usultxt "\t"]
          #set meternum [lindex $usullist 2]
          #set meterden [lindex $usullist 3]
          #set meter "$meternum/$meterden"
          } else {
          set usultxt ""
          puts "no info for $usulname -- $filename"
          }


    set meter $rhythminfo
    set rhythminfolist [split $rhythminfo "/"]
    set meternum [lindex $rhythminfolist 0]
    set meterden [lindex $rhythminfolist 1]

    puts "\tmeter = $meternum/$meterden"

    set factor [expr double($meterden) / double($meternum)]
    #puts "factor = $factor\n" 
    
    

    set inhandle [open $textsrc$fullfilename]
    set nnotes 0
    set nchars 0
    gets $inhandle line

    set counter 0
    set lastbarpos 0.0
    set barnumber 0
    set nmessages 0
    while {![eof $inhandle]} {
        gets $inhandle line
        incr counter
        if {[string length $line] < 10} break
        #puts $line
        set linelist [split $line \t]
        set code [lindex $linelist 1]
        set note [lindex $linelist 2]
        set notenum [lindex $linelist 6]
        set noteden [lindex $linelist 7]

        if {$code == "51"} {
            puts "bar = $barnumber \tmeter = $notenum/$noteden"
            set factor [expr double($noteden) / double($notenum)]
            continue
            } 

        if {$noteden == 0} continue

        set durpercent [lindex $linelist 9]
        set syllable [lindex $linelist 11]
        set barpos [lindex $linelist 12]
        set delta_time [expr $barpos - $lastbarpos]
        set delta_unit [expr floor($barpos) - floor($lastbarpos)]
        set barfrac [expr $barpos - double(floor($barpos))]
        if {$delta_unit && $barfrac > 0.001} {
            puts "expected bar marker prior to $barpos"
            incr nmessages
            }
        set noteduration [expr double($notenum) / double($noteden)]
        set barinc [expr $noteduration * $factor]
        set err [expr $barinc - $delta_time]
        set err [expr abs($err)]
        if {$err > 0.001} {
           puts -nonewline "bar = $barnumber"
           puts " Expecting bar increment of [format "%8.4f" $barinc] instead encountered [format %8.4f $delta_time] at $barpos"
           #set factor [expr $delta_time/$noteduration]
           incr nmessages
           }
        if {[expr floor($barpos)] != [expr floor($lastbarpos)]} {
            incr barnumber
            }
        set lastbarpos $barpos
        if {$nmessages > 10} {puts "aborting....."
            break
            }
        }

    close $inhandle
}



# The main loop starts here.

# Get a list of all the files in the SymbTr2 folder or
# whereever.

cd $textsrc
set filelist [glob  *.txt]
cd ../.


# Process each file until the program crashes or we
# consume the whole list.

set j 1
foreach fullfilename $filelist {
  #to save time ignore file if abc file exists
  set abcname [file rootname $fullfilename].abc
  incr j
  if {$j < 1} continue
  puts -nonewline "$j\t$fullfilename"
  analyze_abc_file $fullfilename
  if {$j > 2300} break
  }
