#tsymb2abc.tcl

# test version.
# The program converts the *.txt files in the folder SymbTr2
# (relative to this script) into abc files and places them in
# the folder abc/. The program requires the following text
# files: makams.txt, forms.txt, usuls.dat, title.dat, composer.dat,
# and tempo.dat to be in the same folder as this script.

#To run this program type:
#tclsh symb2abc.tcl
#followed by <enter>

# textsrc indicates where to find the text files.
global textsrc
set textsrc txt/


set fullfilename "mahur--tavsanca--aksak--aldi_aklim--ii_mahmud.txt"



proc loadmakamtype {} {
    global makamtype
    set inhandle [open makams.txt]
    gets $inhandle line
    gets $inhandle line
    while {![eof $inhandle]} {
        gets $inhandle line
        set linelist [split $line \t]
        set makamname [lindex $linelist 0]
        set makamtype($makamname) $linelist
    }
    close $inhandle
}

proc loadusuls_v3 {} {
    global usulstype
    set inhandle [open usuls_v3.txt]
    gets $inhandle line
    gets $inhandle line
    while {![eof $inhandle]} {
        gets $inhandle line
        set linelist [split $line \t]
        set usulname [lindex $linelist 0]
        set usulstype($usulname) $line
    }
    close $inhandle
}

loadusuls_v3

global makamtype
loadmakamtype

proc loadformtype {} {
    global formtype
    set inhandle [open forms.txt]
    gets $inhandle line
    while {![eof $inhandle]} {
        gets $inhandle line
        set linelist [split $line \t]
        set formname [lindex $linelist 0]
        set formtype($formname) $linelist
    }
    close $inhandle
}

global formtype
loadformtype

proc loadusulinfo {} {
    global usulinfo
    puts "loadusulinfo"
    set inhandle [open usuls.dat]
    while {![eof $inhandle]} {
        gets $inhandle line
        set linelist [split $line \t]
        set filename [lindex $linelist 0]
        #  puts $usulname
        set usulinfo($filename) [lindex $linelist 1]
    }
    close $inhandle
}

proc load_tempoinfo {} {
    global tempoinfo
    set inhandle [open tempo.dat]
    while {![eof $inhandle]} {
        gets $inhandle line
        set linelist [split $line \t]
        set filename [lindex $linelist 0]
        #  puts $usulname
        set tempoinfo($filename) [lindex $linelist 1]
    }
    close $inhandle
}

proc load_titleinfo {} {
    global titleinfo
    set inhandle [open title.dat]
    while {![eof $inhandle]} {
        gets $inhandle line
        set linelist [split $line \t]
        set filename [lindex $linelist 0]
        set titleinfo($filename) [lindex $linelist 1]
    }
    close $inhandle
}

proc load_composerinfo {} {
    global composerinfo
    set inhandle [open composer.dat]
    while {![eof $inhandle]} {
        gets $inhandle line
        set linelist [split $line \t]
        set filename [lindex $linelist 0]
        set composerinfo($filename) [lindex $linelist 1]
    }
    close $inhandle
}

global usulinfo tempoinfo titleinfo composerinfo
loadusulinfo
load_tempoinfo
load_titleinfo
load_composerinfo




# This table indicates how to translate the smaller pitch
# divisions into microtones.

array set ae2abc {#1 ^1/4 #2 ^2/4 #3 ^3/4 #4 ^ #5 ^9/8
    b5 _ b4 _7/8 b3 _3/4 b2 _2/4 b1 _1/4}

set pitchmap {Do C Re D Mi E Fa F Sol G So G La A Si B
do C re D mi E fa F sol G so G la A si B}


# This function converts the input note eg Do4#3 into
# the output for the abc file (eg ^3/4C).
proc ae_note_to_abc {note} {
    global ae2abc
    set k [string index $note 0]
    set octave [string index $note 1]
    set mic [string range $note 2 3]
    #puts "note = $note k = $k octave= $octave mic= $mic"
    if {$k == ""} {return z}
    if {$octave > 4} {set k [string tolower $k]}
    if {$octave > 5} {set k $k'}
    if {$octave < 4} {set k $k,}
    if {$mic != ""} {set k $ae2abc($mic)$k}
    return $k
}


#In order to propagate accidentals (including microtones)
#across a bar we maintain the table activenote($note).
#Note that following music conventions an accidental
#does not propagate across octaves. Thus =Do4 will
#not affect the following Do5 in the bar. On the other
#hand key signature specs apply to all octaves.

#We need to execute reset_localnote at the start of each bar
#to reinstate the key signature specs.

proc reset_localnote {keysignature} {
global activenote
foreach note {A B C D E F G} {
  set activenote($note) $note
  set activenote($note,) $note,
  }
foreach note {a b c d e f g z } {
  set activenote($note) $note
  set activenote($note') $note'
  }
foreach note $keysignature {
  set n [string index $note end]
  set nu [string trimright [string toupper $n] ',]
  set noteu [string toupper $note]
  set activenote($nu) $noteu
  set activenote($nu,) $noteu,
  set nl [string tolower $nu]
  set notel [string tolower $noteu]
  set activenote($nl) $notel
  set activenote($nl') $note'
  }
}

# This function is called for every abc note. Observe
# that we keep all the octave designations. For example 
# C_ C c c' c'' is the pitch of C in different octaves.

proc localnote {note} {
# propagates accidentals across bar
global activenote
regexp -indices \[A-Za-z\] $note i
set i [lindex $i 0]
set n [string range $note $i end]
if {$activenote($n) == $note} {
   #puts "$note i= $i"
   return $n} else {
   set activenote($n) $note
#if no accidentals send a natural note
   if {$i == 0} {return =$note}
   return $note
   }
}


# Divides n1/d1 by n2/d2 and eliminates
# common factors using Euclid's algorithm.
proc fracdivision {n1 d1 n2 d2} {
    global triplet
    set n [expr $n1 * $d2]
    set d [expr $d1 * $n2]
    set m [euclids_algorithm $n $d]
    set n [lindex $m 0]
    set d [lindex $m 1]
    if {[expr $d % 3] == 0} {
        incr triplet
        set d [expr $d/3]
        if {$n > 1} {set n [expr $n/2]
                   } else {set d [expr $d*2]}
        if {$triplet >= 3} {incr triplet -3}
        }
    if {$n == 1 && $d == 1} {return ""}
    if {$d == 1} {return $n}
    if {$d == 2 && $n == 1} {return /}
    return $n/$d
}

proc euclids_algorithm {a b} {
    #eliminate common factor in fraction a/b
    if {$a > $b} {
        set n $a; set m $b} else {
        set n $b; set m $a}
    while {$m != 0} {
        set t [expr $n % $m]
        set n $m
        set m $t
    }
    set a [expr $a/$n]
    set b [expr $b/$n]
    return "$a $b"
}


# decides whether a space should be put between
# two notes if they are in separate beats. A beat
# is assumed to be a quarter note. It uses the
# the barpos given by offset in the text file,
# and the global variable beatwidth which is
# set by the procedure create_abc_file using
# the time signature.

proc beat_boundary {barpos} {
global beatwidth
set beatpos [expr $barpos/$beatwidth]
# is this an integer
set fraction [expr fmod($beatpos,1.0)]
if {$fraction < 0.01} {return 1}
if {$fraction > 0.99} {return 1}
return 0
}


# Determines whether we are at the beginning
# of a new measure (bar) and need to place a bar line.

proc bar_boundary {barpos} {
set fraction [expr fmod($barpos,1.0)]
if {$fraction < 0.01} {return 1}
if {$fraction > 0.99} {return 1}
return 0
}

# Splits the file name into various parts that
# are separated by a double dash --.

proc separate_by_double_dash {name} {
set namelist {}
set i 0
while {[set j [string first -- $name $i]] > -1} {
  set j0 [expr $j - 1]
  set s [string range $name $i $j0]
  lappend namelist $s
  set i [expr $j + 2]
  }
  lappend namelist [string range $name $i end]
  return $namelist
}



proc clean_lyricbar {lyricbar} {
# eliminates unnecessary . and * in the makam lyrics
set lyricbarlist [split $lyricbar " "]
set newlyricbar ""
set count 0
#puts "input lyricbar = $lyricbar"
foreach syl $lyricbarlist {
  if {$syl == "." || $syl ==  "*"} {
    incr count} else {
    if {$count > 0 && $syl != ""} {
      for {set i 0} {$i <$count} {incr i} {
        append newlyricbar " * "
        }
      set count 0
      }
    append newlyricbar " $syl"
   }
  }
  append newlyricbar " | "
#  puts "output newlyricbar = $newlyricbar"
  return $newlyricbar
}



# mcsplit --
#
#   Splits a string based using another string
#
# Arguments:
#   str       string to split into pieces
#   splitStr  substring
#   mc        magic character that must not exist in the orignal string.
#             Defaults to the NULL character.  Must be a single character.
# Results:
#   Returns a list of strings
#
proc mcsplit "str splitStr {mc {\x00}}" {
    return [split [string map [list $splitStr $mc] $str] $mc]
    }




#global filedatainfo


# This function does all the work. Here are a few comments.
# Grace notes have zero duration. We collect them and
# send them out (enclosed in {}} before the next real
# note.
# We have to watch out for triplets. The denominator
# of the duration is divisible by 3.
# Staccato notes are detected by the LNS variable which
# is usually 95 percent except for the notes that are
# staccato. A staccato note is indicated in abc format
# by placing a dot before the note.

# We collect the notes bar by bar in the string
# baroutput. When a bar is filled we append it to
# the string outline. We keep track of the number
# of characters in an abc line and at the end of
# a measure we may start a new line in the output
# abc file (if we already have more than 30 characters).
# We also keep track of the number of notes in the
# system. If we do not have enough notes, then we
# place a backslash at the end of the line to tell
# abcm2ps to continue that system with the information
# in the next line.

# Lyrics are another issue. We cannot have a lyric syllable
# on a rest, so we check for that. We collect the lyric
# syllables in a separate string. We put bar line markers
# in order to avoid losing synchronization with the notes.

set xrefnum 1

proc create_abc_file {fullfilename} {
    global textsrc
    global filedata
    global makamtype
    global pitchmap
    global keysignature
    global usulinfo
    global tempoinfo
    global titleinfo
    global composerinfo
    global triplet
    global beatwidth
    global xrefnum
    global usulstype
 
    set ldenom 16
    set triplet 0
    set l [string length $fullfilename]
    incr l -5
    set filename [string range $fullfilename 0 $l]
    set tempo $tempoinfo($filename)
    set title $titleinfo($filename)
    set composer $composerinfo($filename)

    set filenamelist [separate_by_double_dash $filename]
    set filedata [mcsplit $filename "--"]
    set makamdata [lindex $filedata 0]
    set formdata [lindex $filedata 1]
    set usulname [lindex $filedata 2]
    set rhythminfo $usulinfo($filename)
    if {[info exist usulstype($usulname)]} {
          set usultxt $usulstype($usulname)
          #set usullist [split $usultxt "\t"]
          #set meternum [lindex $usullist 2]
          #set meterden [lindex $usullist 3]
          #set meter "$meternum/$meterden"
          } else {
          set usultxt ""
          puts "no info for $usulname -- $filename"
          }


    set meter $rhythminfo
    set rhythminfolist [split $rhythminfo "/"]
    set meternum [lindex $rhythminfolist 0]
    set meterden [lindex $rhythminfolist 1]

    set makamlist $makamtype($makamdata)
    set makamlist [string map $pitchmap $makamlist]
    
    set keysignature [ae_note_to_abc [lindex $makamlist 1]]
    append keysignature " [ae_note_to_abc [lindex $makamlist 2]]"
    append keysignature " [ae_note_to_abc [lindex $makamlist 3]]"
    append keysignature " [ae_note_to_abc [lindex $makamlist 4]]"
    set keysignature [string map {z ""} $keysignature]
    
    set outhandle [open abc/$filename.abc w]
    puts $outhandle "X: $xrefnum"
    incr xrefnum
    # [SS] 2015-09-04 commaA-E.fmt is now embedded in the header of the
    # abc collection file.
    #puts $outhandle "%%format commaA-E.fmt"
    puts $outhandle "T: $title"
    puts $outhandle "C: $composer"
    puts $outhandle "O: Turkey"
    puts $outhandle "F: http://compmusic.upf.edu/node/140"
    puts $outhandle "N: $filename"
    puts $outhandle "N: makam= $makamdata"
    puts $outhandle "N: form= $formdata"
    puts $outhandle "N: usul= $usultxt"
    #puts $outhandle "M: $rhythminfo"
    puts $outhandle "M: $meter"
    puts $outhandle "L: 1/$ldenom"
    puts $outhandle "K: C $keysignature"
    puts $outhandle "Q: 1/4=$tempo"
    puts $outhandle "%%MIDI tuningsystem comma53"
    
    
    set beatwidth [expr 1.0/$meternum]
    if {$meterden == 8} {set beatwidth [expr $beatwidth * 2.0]}

    set graceseq ""
    set lyricline ""
    set lyricbar ""
    set lyricwords 0
    set lyricwordsinbar 0
    
    set outline ""
    set baroutput ""
    set inhandle [open $textsrc$fullfilename]
    set nnotes 0
    set nchars 0
    gets $inhandle line
    set barkeylist {}

    set counter 0
    reset_localnote $keysignature
    while {![eof $inhandle]} {
        gets $inhandle line
        incr counter
        if {[string length $line] < 10} break
        #puts $line
        set linelist [split $line \t]
        set note [lindex $linelist 2]
        set note [string map $pitchmap $note]
        set notenum [lindex $linelist 6]
        set noteden [lindex $linelist 7]
        set durpercent [lindex $linelist 9]
        set syllable [lindex $linelist 11]
        set barpos [lindex $linelist 12]
        if {$note == "Es"} {set k "z"} else {
           set k [ae_note_to_abc $note]}
        if {$k == "z"} {set syllable ""}
        #puts "\nlocalnote gets $k"
        set k [localnote $k]
        #puts "returns $k"
# staccato ?
        if {$durpercent < 60 && $k != "z"} {set k .$k}

        if {[string length $syllable] > 0} {
         append lyricbar "$syllable "
         incr lyricbarwords
         } else {
         if {$k != "z"} {append lyricbar "* "}
        }

        if {$noteden == 0} {
            append graceseq $k} else {
            #puts "$note $k $notenum/$noteden $triplet"
            set f [fracdivision $notenum $noteden 1 $ldenom]
            if {$triplet == 1} {set k (3$k}
            #puts "$triplet $k$f $barpos"
            if {$graceseq != ""} {
                append baroutput "\{$graceseq\}"
                set graceseq ""
            }
#abcm2ps cannot print notes with length 5 in the numerator
#turn it into a tied note
            if {$notenum == 5} {
                set f [fracdivision 4 $noteden 1 $ldenom]
                append baroutput $k$f-
                incr nnotes
                set f [fracdivision 1 $noteden 1 $ldenom]
                append baroutput $k$f
                incr nnotes
                } else {
                append baroutput $k$f
                incr nnotes
                }
            if {[beat_boundary $barpos] && $triplet == 0} {
                append baroutput " "
                }
            incr l
        }
        
        
# if notenum = 0 this implies a grace note at the beginning of the bar.
# do not reset again because it would erase any accidentals from the grace note
        puts "barpos = $barpos"
        if {[bar_boundary $barpos] == 1 && $notenum != 0} {
             puts "           barpos = $barpos"
             reset_localnote $keysignature
             incr nchars [string length $baroutput]
             append baroutput " | "
             append outline $baroutput
             set lyricbar [clean_lyricbar $lyricbar]
             append lyricline $lyricbar
             set lyricbar ""
             #append lyricline " | "
             set baroutput ""

#       insert line feed if system has more than 20 notes
             if {$nnotes > 20} {
                   append outline "\n"
                   set nnotes 0
                   set nchars 0
                   append outline "w: $lyricline\n"
                   set lyricline ""
                   set lyricwords 0}
                   } elseif {$nchars > 30} {
                   append outline  "\\\n"
                   append outline "w: $lyricline\n"
                   set lyricline ""
                   set lyricwords 0
                   set nchars 0} 
# nchars == 0 signals a line feed was sent
        }
    if {[string index $baroutput end] != "|"} {append baroutput "|"}
    append outline $baroutput\n
    if {$lyricbar != ""} {set lyricbar [clean_lyricbar $lyricbar]
       append lyricline $lyricbar
       append outline "w: $lyricline\n"
       }

    puts $outhandle $outline
    close $inhandle
    close $outhandle
}



# The main loop starts here.

# Get a list of all the files in the SymbTr2 folder or
# whereever.

cd $textsrc

cd ../.


# Process each file until the program crashes or we
# consume the whole list.
set j 88888
set abcname [file rootname $fullfilename].abc
puts "$j\t$fullfilename"
create_abc_file $fullfilename
